=== modified file 'bzrlib/add.py'
--- bzrlib/add.py	2006-08-09 17:25:41 +0000
+++ bzrlib/add.py	2006-08-09 19:53:02 +0000
@@ -79,6 +79,49 @@
         return None
 
 
+class AddFromBaseAction(AddAction):
+    """This class will try to extract file ids from another tree."""
+
+    def __init__(self, base_tree, base_path, to_file=None, should_print=None):
+        super(AddFromBaseAction, self).__init__(to_file=to_file,
+                                                should_print=should_print)
+        self.base_tree = base_tree
+        self.base_path = base_path
+
+    def __call__(self, inv, parent_ie, path, kind):
+        # Place the parent call
+        # Now check to see if we can extract an id for this file
+        file_id, base_path = self._get_base_file_id(path, parent_ie)
+        if file_id is not None:
+            if self.should_print:
+                self._to_file.write('added %s w/ file id from %s\n'
+                                    % (path.raw_path, base_path))
+        else:
+            # we aren't doing anything special, so let the default
+            # reporter happen
+            file_id = super(AddFromBaseAction, self).__call__(
+                        inv, parent_ie, path, kind)
+        return file_id
+
+    def _get_base_file_id(self, path, parent_ie):
+        """Look for a file id in the base branch.
+
+        First, if the base tree has the parent directory,
+        we look for a file with the same name in that directory.
+        Else, we look for an entry in the base tree with the same path.
+        """
+
+        if (parent_ie.file_id in self.base_tree):
+            base_parent_ie = self.base_tree.inventory[parent_ie.file_id]
+            base_child_ie = base_parent_ie.children.get(path.base_path)
+            if base_child_ie is not None:
+                return (base_child_ie.file_id,
+                        self.base_tree.id2path(base_child_ie.file_id))
+        full_base_path = bzrlib.osutils.pathjoin(self.base_path, path.raw_path)
+        # This may return None, but it is our last attempt
+        return self.base_tree.path2id(full_base_path), full_base_path
+
+
 # TODO: jam 20050105 These could be used for compatibility
 #       however, they bind against the current stdout, not the
 #       one which exists at the time they are called, so they

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2006-08-09 16:42:10 +0000
+++ bzrlib/builtins.py	2006-08-09 19:53:02 +0000
@@ -274,18 +274,37 @@
 
     --dry-run will show which files would be added, but not actually 
     add them.
+
+    --file-ids-from will try to use the file ids from the supplied path.
+    It looks up ids trying to find a matching parent directory with the
+    same filename, and then by pure path.
     """
     takes_args = ['file*']
-    takes_options = ['no-recurse', 'dry-run', 'verbose']
+    takes_options = ['no-recurse', 'dry-run', 'verbose',
+                     Option('file-ids-from', type=unicode,
+                            help='Lookup file ids from here')]
     encoding_type = 'replace'
 
-    def run(self, file_list, no_recurse=False, dry_run=False, verbose=False):
+    def run(self, file_list, no_recurse=False, dry_run=False, verbose=False,
+            file_ids_from=None):
         import bzrlib.add
 
-        action = bzrlib.add.AddAction(to_file=self.outf,
-            should_print=(not is_quiet()))
-
-        added, ignored = bzrlib.add.smart_add(file_list, not no_recurse, 
+        if file_ids_from is not None:
+            try:
+                base_tree, base_path = WorkingTree.open_containing(
+                                            file_ids_from)
+            except errors.NoWorkingTree:
+                base_branch, base_path = branch.Branch.open_containing(
+                                            file_ids_from)
+                base_tree = base_branch.basis_tree()
+
+            action = bzrlib.add.AddFromBaseAction(base_tree, base_path,
+                          to_file=self.outf, should_print=(not is_quiet()))
+        else:
+            action = bzrlib.add.AddAction(to_file=self.outf,
+                should_print=(not is_quiet()))
+
+        added, ignored = bzrlib.add.smart_add(file_list, not no_recurse,
                                               action=action, save=not dry_run)
         if len(ignored) > 0:
             if verbose:

=== modified file 'bzrlib/tests/blackbox/test_add.py'
--- bzrlib/tests/blackbox/test_add.py	2006-07-21 04:02:14 +0000
+++ bzrlib/tests/blackbox/test_add.py	2006-08-09 19:53:02 +0000
@@ -136,3 +136,45 @@
         """bzr add foo where foo is missing should error."""
         self.make_branch_and_tree('.')
         self.run_bzr('add', 'missing-file', retcode=3)
+
+    def test_add_from(self):
+        base_tree = self.make_branch_and_tree('base')
+        self.build_tree(['base/a', 'base/b/', 'base/b/c'])
+        base_tree.add(['a', 'b', 'b/c'])
+        base_tree.commit('foo')
+
+        new_tree = self.make_branch_and_tree('new')
+        self.build_tree(['new/a', 'new/b/', 'new/b/c', 'd'])
+
+        os.chdir('new')
+        out, err = self.run_bzr('add', '--file-ids-from', '../base')
+        self.assertEqual('', err)
+        self.assertEqualDiff('added a w/ file id from a\n'
+                             'added b w/ file id from b\n'
+                             'added b/c w/ file id from b/c\n',
+                             out)
+
+        new_tree.read_working_inventory()
+        self.assertEqual(base_tree.path2id('a'), new_tree.path2id('a'))
+        self.assertEqual(base_tree.path2id('b'), new_tree.path2id('b'))
+        self.assertEqual(base_tree.path2id('b/c'), new_tree.path2id('b/c'))
+
+    def test_add_from_subdir(self):
+        base_tree = self.make_branch_and_tree('base')
+        self.build_tree(['base/a', 'base/b/', 'base/b/c', 'base/b/d'])
+        base_tree.add(['a', 'b', 'b/c', 'b/d'])
+        base_tree.commit('foo')
+
+        new_tree = self.make_branch_and_tree('new')
+        self.build_tree(['new/c', 'new/d'])
+
+        os.chdir('new')
+        out, err = self.run_bzr('add', '--file-ids-from', '../base/b')
+        self.assertEqual('', err)
+        self.assertEqualDiff('added c w/ file id from b/c\n'
+                             'added d w/ file id from b/d\n',
+                             out)
+
+        new_tree.read_working_inventory()
+        self.assertEqual(base_tree.path2id('b/c'), new_tree.path2id('c'))
+        self.assertEqual(base_tree.path2id('b/d'), new_tree.path2id('d'))

=== modified file 'bzrlib/tests/test_smart_add.py'
--- bzrlib/tests/test_smart_add.py	2006-08-09 17:25:41 +0000
+++ bzrlib/tests/test_smart_add.py	2006-08-09 19:53:02 +0000
@@ -19,7 +19,12 @@
 import unittest
 
 from bzrlib import errors, ignores, osutils
-from bzrlib.add import smart_add, smart_add_tree, AddAction
+from bzrlib.add import (
+    AddAction,
+    AddFromBaseAction,
+    smart_add,
+    smart_add_tree,
+    )
 from bzrlib.tests import TestCase, TestCaseWithTransport, TestSkipped
 from bzrlib.errors import NoSuchFile
 from bzrlib.inventory import InventoryFile, Inventory
@@ -145,7 +150,7 @@
         self.assertSubset(('inertiatic/foo.pyc',), ignored['*.py[co]'])
 
 
-class CustomIDAddAction(AddAction):
+class AddCustomIDAction(AddAction):
 
     def __call__(self, inv, parent_ie, path, kind):
         # The first part just logs if appropriate
@@ -242,7 +247,7 @@
 
     def test_custom_ids(self):
         sio = StringIO()
-        action = CustomIDAddAction(to_file=sio, should_print=True)
+        action = AddCustomIDAction(to_file=sio, should_print=True)
         self.build_tree(['file1', 'dir1/', 'dir1/file2'])
 
         wt = self.make_branch_and_tree('.')
@@ -259,6 +264,109 @@
                                 in wt.inventory.iter_entries()])
 
 
+class TestAddFrom(TestCaseWithTransport):
+    """Tests for AddFromBaseAction"""
+
+    def make_base_tree(self):
+        self.base_tree = self.make_branch_and_tree('base')
+        self.build_tree(['base/a', 'base/b',
+                         'base/dir/', 'base/dir/a',
+                         'base/dir/subdir/',
+                         'base/dir/subdir/b',
+                        ])
+        self.base_tree.add(['a', 'b', 'dir', 'dir/a',
+                            'dir/subdir', 'dir/subdir/b'])
+        self.base_tree.commit('creating initial tree.')
+
+    def add_helper(self, base_tree, base_path, new_tree, file_list,
+                   should_print=False):
+        to_file = StringIO()
+        base_tree.lock_read()
+        try:
+            new_tree.lock_write()
+            try:
+                action = AddFromBaseAction(base_tree, base_path,
+                                           to_file=to_file,
+                                           should_print=should_print)
+                smart_add_tree(new_tree, file_list, action=action)
+            finally:
+                new_tree.unlock()
+        finally:
+            base_tree.unlock()
+        return to_file.getvalue()
+
+    def test_copy_all(self):
+        self.make_base_tree()
+        new_tree = self.make_branch_and_tree('new')
+        files = ['a', 'b',
+                 'dir/', 'dir/a',
+                 'dir/subdir/',
+                 'dir/subdir/b',
+                ]
+        self.build_tree(['new/' + fn for fn in files])
+        self.add_helper(self.base_tree, '', new_tree, ['new'])
+
+        for fn in files:
+            base_file_id = self.base_tree.path2id(fn)
+            new_file_id = new_tree.path2id(fn)
+            self.assertEqual(base_file_id, new_file_id)
+
+    def test_copy_from_dir(self):
+        self.make_base_tree()
+        new_tree = self.make_branch_and_tree('new')
+
+        self.build_tree(['new/a', 'new/b', 'new/c',
+                         'new/subdir/', 'new/subdir/b', 'new/subdir/d'])
+        self.add_helper(self.base_tree, 'dir', new_tree, ['new'])
+
+        # We 'a' and 'b' exist in the root, and they are being added
+        # in a new 'root'. Since ROOT ids are not unique, we will
+        # use those ids
+        # TODO: This will probably change once trees have a unique root id
+        # It is definitely arguable that 'a' should get the id of
+        # 'dir/a' not of 'a'.
+        self.assertEqual(self.base_tree.path2id('a'),
+                         new_tree.path2id('a'))
+        self.assertEqual(self.base_tree.path2id('b'),
+                         new_tree.path2id('b'))
+
+        # Because we specified 'dir/' as the base path, and we have
+        # nothing named 'subdir' in the base tree, we should grab the
+        # ids from there
+        self.assertEqual(self.base_tree.path2id('dir/subdir'),
+                         new_tree.path2id('subdir'))
+        self.assertEqual(self.base_tree.path2id('dir/subdir/b'),
+                         new_tree.path2id('subdir/b'))
+
+        # These should get newly generated ids
+        c_id = new_tree.path2id('c')
+        self.assertNotEqual(None, c_id)
+        self.failIf(c_id in self.base_tree)
+
+        d_id = new_tree.path2id('subdir/d')
+        self.assertNotEqual(None, d_id)
+        self.failIf(d_id in self.base_tree)
+
+    def test_copy_existing_dir(self):
+        self.make_base_tree()
+        new_tree = self.make_branch_and_tree('new')
+        self.build_tree(['new/subby/', 'new/subby/a', 'new/subby/b'])
+
+        subdir_file_id = self.base_tree.path2id('dir/subdir')
+        new_tree.add(['subby'], [subdir_file_id])
+        self.add_helper(self.base_tree, '', new_tree, ['new'])
+        # Because 'subby' already points to subdir, we should add
+        # 'b' with the same id
+        self.assertEqual(self.base_tree.path2id('dir/subdir/b'),
+                         new_tree.path2id('subby/b'))
+
+        # 'subby/a' should be added with a new id because there is no
+        # matching path or child of 'subby'.
+        a_id = new_tree.path2id('subby/a')
+        self.assertNotEqual(None, a_id)
+        self.failIf(a_id in self.base_tree)
+
+
 class TestAddNonNormalized(TestCaseWithTransport):
 
     def make(self):

